"""
Implementation of models concept on top of Google cloud firestore.
Firestore entities try to make interaction with firestore as simple as possible for the developer.
"""

from firestore.entity import Entity
from firestore.db import SERVER_TIMESTAMP